#!/usr/bin/env python
# -*- encoding : ascii -*-
#
# The main program of PyClonezilla
#
# Copyright (C) 2010~ Zong-shuo Jheng <c00aso00 at nchc org tw>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import pygtk
pygtk.require('2.0')
import os, sys
import locale, gettext

from stage import *
from util import *

if sys.version_info < (2, 6):
        print "must use python 2.6 or greater"
	pyst = 1
else:
	pyst = 0

try:
	import gtk
except:
	print >> sys.stderr, "Can't import python gtk"
	sys.exit(1)

try:
	import gobject
except:
	print >> sys.stderr, "Can't import python gobject"
	sys.exit(1)

try:
	import vte
except:
	print >> sys.stderr, "Can't import python vte"
	sys.exit(1)

_ = lang.gettext



#========== Querying Dailog Group ==========
def query_yes_no(msg, parent=None):
    """
    ask for yes or no.

    @param msg str message string.
    @param parent
    @return True if the respone is yes.
    """
    return query(msg, \
                 gtk.MESSAGE_QUESTION, \
                 gtk.BUTTONS_YES_NO,\
                 gtk.RESPONSE_YES)

def query_confirm(msg, parent=None):
    """
    ask for cofirm.

    @param msg str message string.
    @param parent
    @return True if the respone is ok.
    """
    return query(msg,\
                 gtk.MESSAGE_QUESTION, \
                 gtk.BUTTONS_OK_CANCEL,\
                 gtk.RESPONSE_YES)

def query(msg, msg_type, button_type, assert_res=None, parent=None):
    dlg = gtk.MessageDialog \
        (parent, gtk.DIALOG_MODAL, \
         msg_type, button_type, msg)
    ret = dlg.run ()
    dlg.destroy ()

    if assert_res is None:
        return ret
    else:
        return ret == assert_res

def show_error(msg, title=None, parent=None):
    """
    display error message.

    @param title str the title in dialog.
    @parant
    """
    dlg = gtk.MessageDialog \
            (parent, gtk.DIALOG_MODAL, \
            gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, msg)

    if title:
        dlg.set_title(title)

    dlg.run()
    dlg.destroy()


#======================
# Main Window Class 
#======================
class MainWin:
    #===========================
    # Initialization Function 
    #===========================
    def __init__(self):
        win=gtk.Window(gtk.WINDOW_TOPLEVEL)
        win.set_title(_('''Clonezilla GUI (Free Software Lab at the NCHC, Taiwan)'''))
        win.set_resizable(False)

        if os.path.exists("pics/clonezilla.jpg"):
            win.set_icon_from_file("pics/clonezilla.jpg")

        win.connect('delete_event', self.on_close)

        # ===================================================================================
        # Setup the APP backbround image.
        # Suck PyGtk set background http://faq.pygtk.org/index.py?req=show&file=faq04.016.htp
        # ===================================================================================
        pixbuf = gtk.gdk.pixbuf_new_from_file('pics/bg.PNG')
        pixmap, mask = pixbuf.render_pixmap_and_mask()
        width, height = pixmap.get_size()
        del pixbuf

        evbox = gtk.EventBox()
        win.add(evbox)

        evbox.set_app_paintable(True)
        win.resize(width, height)
        evbox.realize()
        evbox.window.set_back_pixmap(pixmap, False)
        del pixmap

        evbox.show()

        self.MainVBox=gtk.VBox(False, 3)
        evbox.add(self.MainVBox)

        #================================================================= 
        # The two stacks that store the stage windows and titles will be
        # pushed and poped while being triggled by self.on_go_forward() 
        # and self.on_go_backward(), respectively. 
        #=================================================================
        self.stage_stack = []
        self.stage_title_stack = []

        #====================================
        # Upper parts: Header
        #   To setup the title of each stages.
        #====================================
        self.FixedHeader = gtk.Fixed()

        self.HeaderText = gtk.Label()
        self.HeaderText.show()

        self.FixedHeader.put(self.HeaderText, 20, 20)
        self.FixedHeader.show_all()

        self.MainVBox.pack_start(self.FixedHeader, False, True, 0)

        #====================================
        # Middle parts: main GUI 
        #====================================
        self.view=gtk.Viewport()
        self.view.set_size_request(460,365)
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        self.scroll=gtk.ScrolledWindow()
        self.scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scroll.set_shadow_type(gtk.SHADOW_IN)

        self.Page = Stage1_Clonezilla_or_Shell_Page(win, self)
        self.set_header_text(self.Page.get_title())

        # Push initial stage to stack. 
        self.stage_stack.append(self.Page)
        self.stage_title_stack.append(self.Page.get_title())

        self.scroll.add(self.Page.get_widget())
        self.scroll.show_all()

        self.view.add(self.scroll)

        self.FixedMainFrame = gtk.Fixed()

        self.FixedMainFrame.put(self.view, 165 , 35)
        self.FixedMainFrame.show_all()

        self.MainVBox.pack_start(self.FixedMainFrame, False, True, 0)

        #====================================
        # Lower parts: buttons at bottom
        #====================================
        self.LowerHBox = gtk.HBox(False,2)

        btn=gtk.Button(stock=gtk.STOCK_ABOUT)
        #btn.connect('clicked', self.on_about)
        btn.show()
        self.LowerHBox.pack_start(btn, False, True, 2)

        btn=gtk.Button(stock=gtk.STOCK_GO_FORWARD)
        btn.connect('clicked', self.on_go_forward)
        btn.show()
        self.LowerHBox.pack_end(btn, False, True, 5)
        self.go_forward_btn=btn

        btn=gtk.Button(stock=gtk.STOCK_GO_BACK)
        btn.connect('clicked', self.on_go_backward)
        btn.hide()
        self.LowerHBox.pack_end(btn, False, True, 5)
        self.go_back_btn=btn

        btn=gtk.Button(stock=gtk.STOCK_CLOSE)
        btn.connect('clicked', self.on_cancel)
        btn.show()
        self.LowerHBox.pack_end(btn, False, True, 5)
        self.cancel_btn=btn

        self.LowerHBox.show()
        self.MainVBox.pack_end(self.LowerHBox, False, True, 2)


        self.MainVBox.show()

        win.set_default_size( 640, 480 )
        win.show()

        self.win=win

        self.pid=-1
        self.complete=False

    #====================================================
    # Unmount /home/partimag if exsit and being mounted.
    #====================================================
    def umount_home_partimag_and_close(self):
        if os.path.exists("/home/partimag"):
            if os.path.ismount("/home/partimag"):
                try:
                    subprocess.Popen(["sudo", "umount", "/home/partimag" ], stdout=subprocess.PIPE)
                except OSError:
                    print str(OSError)

        gtk.main_quit()

    #====================================
    # Confirm to exit the main program 
    #====================================
    def confirm_close(self):
        if not self.Page.process_finished:
            dialog = gtk.Dialog( "",
                                 self.win,
                                 gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,                       
                                 (gtk.STOCK_OK, gtk.RESPONSE_YES, gtk.STOCK_CANCEL, gtk.RESPONSE_NO))
            hbox = gtk.HBox(False, 2)
            image = gtk.Image()
            image.set_padding(0, 0)
            icon = gtk.STOCK_DIALOG_QUESTION
            image.set_from_stock(icon, gtk.ICON_SIZE_DIALOG)
            image.show()
            hbox.pack_start(image)
            hbox.pack_start(gtk.Label(_("Are you sure to exit the program now?")),True,True,15)
            hbox.show_all()

            dialog.vbox.pack_start(hbox,True,True,15)
            dialog.show_all()

            result = dialog.run()
            if result != gtk.RESPONSE_YES:
                dialog.destroy()
                return 0

        self.umount_home_partimag_and_close()

    #===============================
    # Closing Window 
    #===============================
    def on_close(self, w, evt):
        return self.confirm_close()

    #===============================
    # Canceling The Application
    #===============================
    def on_cancel(self, btn):
        return self.confirm_close()

    #====================================
    # Go Forward To The Next Stage
    #====================================
    def on_go_forward(self, btn):
        get_return = self.Page.go_next_page(self)
        if get_return != -1:
            self.scroll.remove(self.Page.view)
            self.Page = get_return
            self.Page.default_button_Setting(self)
            self.stage_stack.append(self.Page)

            self.stage_title_stack.append(self.Page.get_title())
            self.set_header_text(self.Page.get_title())

            self.scroll.add(self.Page.get_widget())

            if self.Page.title == _("Processing"):
                self.Page.run_process()

    #==================================== 
    # Go Backward To The Previous Stage
    #====================================
    def on_go_backward(self, btn):
        if len(self.stage_stack) > 1:
            self.scroll.remove(self.Page.get_widget())
            self.stage_stack.pop()
            self.Page = self.stage_stack[len(self.stage_stack) - 1]
            self.Page.default_button_Setting(self)

            self.stage_title_stack.pop()
            self.set_header_text(self.stage_title_stack[len(self.stage_title_stack) - 1])

            self.scroll.add(self.Page.get_widget()) 

    def get_widget(self):
        return self.scroll

    def set_header_text(self, text):
        self.HeaderText.set_text(text)

    #=============================================
    # Setting the state of buttons in the bottom.
    #=============================================
    def set_button(self, button, mode):
        if button == "forward":
            if mode == "show":
                self.go_forward_btn.show()
            elif mode == "hide":
                self.go_forward_btn.hide()
            elif mode == "active":
                self.go_forward_btn.set_sensitive(True)
            else:
                self.go_forward_btn.set_sensitive(False)
        elif button == "backward":
            if mode == "show":
                self.go_back_btn.show()
            elif mode == "hide":
                self.go_back_btn.hide()
            elif mode == "active":
                self.go_back_btn.set_sensitive(True)
            else:
                self.go_back_btn.set_sensitive(False)
        else:  #button == "cancel"
            if mode == "active":
                self.cancel_btn.set_sensitive(True)
            else:
                self.cancel_btn.set_sensitive(False)
pass

#===================
# Main Entry Point 
#===================
def startgui():
    """
    Launch the application.
    """
    MainWin()
    gtk.main()


if __name__ =="__main__":
    startgui()


