#!/usr/bin/env python
# -*- encoding : ascii -*-
#
# Stage classes for the main window of Clonezilla GUI
#
# Copyright (c) 2010~ Zong-shuo Jheng <c00aso00 at nchc org tw>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import pygtk
pygtk.require('2.0')
import os, sys
import string, datetime
import gtk, gobject, vte
import locale, gettext
import threading
import time


import commands

from util import *
from thread import *

_ = lang.gettext

All_Arguments = { 'Source'        : '',   # The source device or image file
                  'Source Type'   : '',   # Type of the source ((DEVICE / PARTITION) | IMAGE)
                  'Target'        : '',   # The target device or image file
                  'Target Type'   : '',   # Type of the target ((DEVICE / PARTITION) | IMAGE)
                  'Operation'     : '',   # Clone | restore | onthfly
                 }

clone_priority_opts =     [ ["q2", _("Priority: partclone > partimage > dd"), True],
                            ["q1", _("Priority: Only dd (supports all filesystem, but inefficient)"), False],
                            ["q", _("Priority: ntfsclone > partimage > dd"), False],
                            ["null", _("Priority: partimage > dd(no ntfsclone)"), False] ]

clone_mess_opts =        [ ["c", _("Client waits for confirmation before cloneing"), True],
                           ["j2", _("Clone the hidden data between MBR and 1st partition"), True],
                           ["a", _("Do NOT force to turn on HD DMA"), False],
                           ["rm-win-swap-hib", _("Remove page and hibernation files in Win if exists"), False],
                           ["ntfs-ok", _("Skip checking NTFS integrity, even bad sectors (ntfsclone only)"), False],
                           ["rescure", _("Continue reading next one when disk blocks read errors"), False],
                           ["fsck-src-part", _("Check and repair source file system before saving"), False],
                           ["gm", _("Generate image MD5 checksums"), False],
                           ["gs", _("Generate image SHA1 checksums"), False] ]

clone_compression_opts = [ ["z1", _("gzip compression (fast with a smaller image)"), True],
                           ["z2", _("bzip2 compression (slowest but smallest image)"), False],
                           ["z3", _("lzo compression (faster with image size approx. to that of gzip)(NOTE!!)"), False],
                           ["z4", _("lzma compression (slowest but also small image, faster decompression than bzip2)"), False],
                           ["z5", _("xz compression (slowest but also small image, faster decompression than bzip2)"), False],
                           ["z6", _("lzip compression (slowest but also small image, faster decompression than bzip2)"), False],
                           ["z0", _("No compression (fastest but largest image size)"), False] ]

restore_partition_opts = [ ["null", _("Use the partition table from the image"), True],
                           ["k", _("Do NOT create a partition table on the target disk"), False],
                           ["k1", _("Create partition table proportionally (OK for MBR format, not GPT)"), False],
                           ["j0", _("Use dd to create partition (NOT OK if logical drives exits)"), False] ]
                           #["k2", _("Enter command line prompot to create partition manually later"), False],

restore_adv_opts =       [ ["g auto", _("Reinstall grub in client disk MBR (only if grub config exists)"), True],
                           ["e1 auto", _("Automatically adjust filesystem geometry for a NTFS boot partition if exists"), True],
                           ["e2", _("sfdisk uses CHS of hard drive from EDD(for non-grub boot loader)"), True],
                           ["hn0 PC", _("Change MS Win hostname (based on IP address) after clone"), False],
                           ["hn1 PC", _("Change MS Win hostname (based on MAC address) after clone"), False],
                           ["b", _("Run clone in batch mode (DANGEROUS!)"), False],
                           ["c", _("Client waits for confirmation before cloneing"), True],
                           ["t", _("Client does not restore the MBR (Master Boot Record)"), False],
                           ["t1", _("Client restores the prebuilt MBR from syslinux (For Windows only)"), False],
                           ["r", _("Try to resize the filesystem to fit partition size"), False],
                           ["e", _("sfdisk uses the CHS value of hard drive from the saved image"), False],
                           ["icrc", _("Ignore CRC checking of partclone"), False],
                           ["j1", _("Write MBR (512 B) again after image is restored. Not OK for partition table differ"), False],
                           ["j2", _("Clone the hidden data between MBR and 1st partition"), True],
                           ["cm", _("Check image by MD5 checksums"), False],
                           ["cs", _("Check image by SHA1 checksums"), False],
                           ["a", _("Do NOT force to turn on HD DMA"), False],
                           ["o0", _("Run script in $OCS_PRERUN_DIR before clone starts"), False], 
                           ["o1", _("Run script in $OCS_POSTRUN_DIR as clone finishes"), False] ]

onthefly_adv_opts1 =     [ ["g auto", _("Reinstall grub on target hard disk boot sector"), True],
                           ["e1 auto", _("Automatically adjust filesystem geometry for a NTFS boot partition if exists"), True],
                           ["e2", _("sfdisk uses CHS of hard drive from EDD(for non-grub boot loader)"), True],
                           ["j2", _("Clone the hidden data between MBR and 1st partition"), True],
                           ["r", _("Resize the filesystem to fit partition size of target partition"), False],
                           ["m", _("DO NOT clone boot loader"), False],
                           ["rescue", _("Continue reading next one when disk blocks read errors"), False],
                           ["fsck-src-part", _("Check and repair source file system before cloning"), False],
                           ["o", _("Force to load the saved HD CHS value"), False],
                           ["b", _("Run clone in batch mode (DANGEROUS!)"), False],  ]
                           #["v", _("Show verbose information"), False]

onthefly_adv_opts2 =     [ ["null", _("Use the partition table from the source disk"), True],
                           ["k", _("Do NOT create a partition table on the target disk"), False],
                           ["k1", _("Create partition table proportionally (OK for MBR format, not GPT)"), False], ]
                           #["k2", _("Enter command line prompot to create partition manually later"), False]


finished__power_opts =   [ ["p true", _("Do nothing when the clone finishes"), True],
                           ["p reboot", _("Reboot client when the clone finishes"), False],
                           ["p poweroff", _("Shutdown client when the clone finishes"), False] ]

#========== Entrance Stage Page Class ==========
class Stage1_Clonezilla_or_Shell_Page:
    def __init__(self, win, parent):
        self.win = win
        self.parent = parent
        self.title = _("Welcome to use PyClonezilla!")
        self.process_finished = False

        self.view=gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        box = gtk.VBox(False, 0)
        self.view.add(box)
        box.show()

        self.RadioButton1 = gtk.RadioButton(None , _("Start Clonezilla"))
        box.pack_start(self.RadioButton1, False, True, 0)
        self.RadioButton1.show()

        self.RadioButton2 = gtk.RadioButton(self.RadioButton1 , _("Enter Shell"))
        box.pack_start(self.RadioButton2, False, True, 0)
        self.RadioButton2.show()

        self.view.show_all()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="active")
        parent.set_button(button="forward",mode="show")
        parent.set_button(button="backward",mode="hide")
        parent.set_button(button="cancel",mode="active")

    def go_next_page(self, parent):
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")

        clonezilla = self.RadioButton1.get_active()
        if clonezilla:
            return Stage2_Select_Source(self.win,self.parent)
        else:
            return Enter_Shell(self.parent)

pass

#========== Entrance Stage Page Class ==========
class Enter_Shell:
    def __init__(self, parent):
        self.view = gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        self.parent = parent
        self.title = _("Terminal Mode")
        self.process_finished = False

        vt = vte.Terminal()
        vt.fork_command()

        self.view.add(vt)
        self.view.show_all()

        #vt.feed_child('/opt/drbl/sbin/ocs-sr -q2 -c -j2 -z1 -i 1000000 -p true saveparts test-img sda1\n')

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="hide")
        parent.set_button(button="backward",mode="active")
        parent.set_button(button="backward",mode="show")
        parent.set_button(button="cancel",mode="active")

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

pass

#===============================================

class Stage2_Select_Source:
    def __init__(self, win, parent):
        self.win = win
        self.parent = parent
        self.title = _("Select a device partition or image file as the source:")
        self.process_finished = False

        self.page = Select_Source_or_Target_Page(self.win, "Select Source", self.parent)

        self.page.set_radiobutton(True, True)
        self.view = self.page.get_page()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="show")
        if self.page.selection == "":
            parent.set_button(button="forward",mode="inactive")
        else:
            parent.set_button(button="forward",mode="active")

        parent.set_button(button="backward",mode="show")
        parent.set_button(button="backward",mode="active")
        parent.set_button(button="cancel",mode="active")

    def go_next_page(self, parent):
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="backward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")
        if self.page.get_radiobutton():        # return True while DEVICE or PARTITION being selected.
            (model, paths) = self.page.tree.get_selection().get_selected()
            if paths and len(model.get_path(paths)) < 2:
                All_Arguments['Source Type'] = "DEVICE"
            else:
                All_Arguments['Source Type'] = "PARTITION"

            All_Arguments['Source'] = self.page.get_selection()
        else:
            All_Arguments['Source Type'] = "IMAGE"
            (model, paths) = self.page.tree.get_selection().get_selected()

            # If the network source works, this option should be modified.
            if not paths or len(model.get_path(paths)) < 3:
                text = _("Your selected image file path may be invalid.")
                md = gtk.MessageDialog(self.win, 
                                       gtk.DIALOG_DESTROY_WITH_PARENT, 
                                       gtk.MESSAGE_ERROR, 
                                       gtk.BUTTONS_CLOSE, 
                                       text )
                md.run()
                md.destroy()

                parent.set_button(button="forward",mode="active")
                parent.set_button(button="backward",mode="active")
                parent.set_button(button="cancel",mode="active")
                return -1
  
            All_Arguments['Source'] = model.get_value(paths, 1)

        return Stage3_Select_Target(self.win, self.page.get_radiobutton(), self.parent)
pass

#===============================================

class Stage3_Select_Target:
    def __init__(self, win, source, parent):
        self.win = win
        self.parent = parent
        self.title = _("Select a device partition or image file as the target:")
        self.process_finished = False

        self.page = Select_Source_or_Target_Page(self.win, "Select Target", self.parent)
 
        if source:
            self.page.set_radiobutton(True, True)
        else:
            # while False, the target should not be image files. 
            self.page.set_radiobutton(True, False)

        self.view = self.page.get_page()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="show")
        if self.page.selection == "":
            parent.set_button(button="forward",mode="inactive")
        else:
            parent.set_button(button="forward",mode="active")

        parent.set_button(button="backward",mode="show")
        parent.set_button(button="backward",mode="active")
        parent.set_button(button="cancel",mode="active")

    def go_next_page(self, parent):
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="backward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")

        if self.page.get_selection() == "":
            text = _("You haven't select a partition or image file as the target yet, please select one.")
            md = gtk.MessageDialog(self.win, 
                                   gtk.DIALOG_DESTROY_WITH_PARENT, 
                                   gtk.MESSAGE_ERROR, 
                                   gtk.BUTTONS_CLOSE, 
                                   text )
            md.run()
            md.destroy()
            parent.set_button(button="forward",mode="active")
            parent.set_button(button="backward",mode="active")
            parent.set_button(button="cancel",mode="active")
            return -1
        else:
            if self.page.RadioButton2.get_active():
                All_Arguments['Target Type'] = "IMAGE"
            else:
                (model, paths) = self.page.tree.get_selection().get_selected()
                if len(model.get_path(paths)) < 2:
                    All_Arguments['Target Type'] = "DEVICE"
                else:
                    All_Arguments['Target Type'] = "PARTITION"

            if All_Arguments['Source Type'] == "DEVICE":
                if All_Arguments['Target Type'] == "IMAGE":
                    All_Arguments['Operation'] = "savedisk"
                else:
                    All_Arguments['Operation'] = "onthefly"
            elif All_Arguments['Source Type'] == "PARTITION":
                if All_Arguments['Target Type'] == "IMAGE":
                    All_Arguments['Operation'] = "saveparts"
                else:
                    All_Arguments['Operation'] = "onthefly"
            else:
                if All_Arguments['Target Type'] == "DEVICE":
                    All_Arguments['Operation'] = "restoredisk"
                else:
                    All_Arguments['Operation'] = "restoreparts"

            All_Arguments['Target'] = self.page.get_selection()
            return Stage4_Advance_Options(self.win,self.parent)

pass

#===============================================

class Stage4_Advance_Options:
    def __init__(self, win, parent):
        self.win = win
        self.parent = parent
        self.title = _("Options:")
        self.process_finished = False

        self.opt_flags = [False, False, False]
        self.opt_stack = []

        self.view = gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        self.mode = ""

        if All_Arguments['Source Type'] == "IMAGE":
            self.mode = "Restore"
            self.view.add(self.load_opts("Restore"))
        else:
            if All_Arguments['Operation'] == "onthefly":
                self.mode = "Onthefly"
                self.view.add(self.load_opts("Onthefly"))
            else:
                self.mode = "Clone"
                self.view.add(self.load_opts("Clone"))

        self.view.show()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="show")
        parent.set_button(button="forward",mode="active")
        parent.set_button(button="backward",mode="show")
        parent.set_button(button="backward",mode="active")
        parent.set_button(button="cancel",mode="active")

    def go_next_page(self, parent):
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="backward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")

        opts = []
        operation_cmd = ""
        operation_cmd2 = ""

        if All_Arguments['Operation'] == "onthefly":
            op = _("OnTheFly")
            opts = onthefly_adv_opts1 + onthefly_adv_opts2

            operation_cmd += "/opt/drbl/sbin/ocs-onthefly"
            operation_cmd2 += " -f %s -t %s" % (All_Arguments['Source'].split("/")[-1], All_Arguments['Target'].split("/")[-1])

        elif All_Arguments['Operation'] == "savedisk" or All_Arguments['Operation'] == "savepart":
            op = _("Clone")
            opts = clone_priority_opts + clone_mess_opts + clone_compression_opts

            operation_cmd += "/opt/drbl/sbin/ocs-sr"
            operation_cmd2 += " -p true %s %s %s" % (All_Arguments['Operation'], All_Arguments['Target'], All_Arguments['Source'].split("/")[-1])

        else:
            op = _("Restore")
            opts = restore_partition_opts + restore_adv_opts

            operation_cmd += "/opt/drbl/sbin/ocs-sr"
            #operation_cmd2 += " -f %s -t %s" % (All_Arguments['Source'].split("::")[1], All_Arguments['Target'].split("::")[1])


        for opt in opts:
            if opt[2] == True:
                if opt[0] != "null":
                    operation_cmd += " -%s" % opt[0]

        operation_cmd += " -nogui"
        operation_cmd += operation_cmd2

        text = _("Are you sure to do \"%s\" operation from %s to %s? ") % \
                 (op, All_Arguments['Source'].split("::")[1], All_Arguments['Target'].split("::")[1])

        dialog = gtk.Dialog( "",
                             self.win,
                             gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,                       
                             (gtk.STOCK_OK, gtk.RESPONSE_YES, gtk.STOCK_CANCEL, gtk.RESPONSE_NO))
        hbox = gtk.HBox(False, 2)
        image = gtk.Image()
        image.set_padding(0, 0)
        icon = gtk.STOCK_DIALOG_WARNING
        image.set_from_stock(icon, gtk.ICON_SIZE_DIALOG)
        image.show()
        hbox.pack_start(image)
        hbox.pack_start(gtk.Label(text),True,True,15)
        hbox.show_all()
        dialog.vbox.pack_start(hbox,True,True,15)

        hbox2 = gtk.HBox(False, 2)
        lbl = gtk.Label(operation_cmd)
        hbox2.pack_start(lbl)
        hbox2.show_all()
        dialog.vbox.pack_start(hbox2,True,True,15)

        dialog.show_all()
        result = dialog.run()

        if result == gtk.RESPONSE_YES:
            dialog.destroy()
            dialog2 = gtk.Dialog( "",
                                  self.win,
                                  gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,                       
                                  (gtk.STOCK_OK, gtk.RESPONSE_YES, gtk.STOCK_CANCEL, gtk.RESPONSE_NO))
            hbox = gtk.HBox(False, 2)
            image = gtk.Image()
            image.set_padding(0, 0)
            icon = gtk.STOCK_DIALOG_WARNING
            image.set_from_stock(icon, gtk.ICON_SIZE_DIALOG)
            image.show()
            hbox.pack_start(image)
            hbox.pack_start(gtk.Label(_("Are you sure to do this operation? After you press \"YES\", it can NOT be rollbacked!")),True,True,15)
            hbox.show_all()
            dialog2.vbox.pack_start(hbox,True,True,15)

            result = dialog2.run()
            if result == gtk.RESPONSE_YES:
                dialog2.destroy()
                return Stage5_Executing_Clonezilla_Core(self.win, self.parent, operation_cmd)
            else:
                dialog2.destroy()
                parent.set_button(button="forward",mode="active")
                parent.set_button(button="backward",mode="active")
                parent.set_button(button="cancel",mode="active")
                return -1
        else:
            dialog.destroy()
            parent.set_button(button="forward",mode="active")
            parent.set_button(button="backward",mode="active")
            parent.set_button(button="cancel",mode="active")
            return -1

    def on_radiobutton_toggled(self, widget, opts, index): 
        if widget.get_active():
            opts[index][2] = True
        else:
            opts[index][2] = False

    def on_button_clicked(self, widget, index, subVbox):
        for idx in range(len(self.opt_stack)):
            if index == idx:
                if self.opt_flags[idx] == True:
                    self.opt_flags[idx] = False
                    self.opt_stack[idx].hide()
                else:
                    self.opt_flags[idx] = True
                    self.opt_stack[idx].show_all()
            else:
                self.opt_flags[idx] = False
                self.opt_stack[idx].hide()

    def load_opts(self, mode):
        vbox = gtk.VBox(False, 7)
        vbox.show()

        hbox = gtk.HBox(False, 2)
        label = gtk.Label(_("If you have no idea, keep the default value and do NOT \nchange anything.\n"))
        label.show()
        hbox.pack_start(label, False, False, 10)

        image = gtk.Image()
        image.set_padding(0, 0)
        icon = gtk.STOCK_INFO
        image.set_from_stock(icon, gtk.ICON_SIZE_BUTTON)
        image.show()
        button = gtk.Button()
        button.add(image)
        button.set_relief(gtk.RELIEF_NONE)
        button.show()
        #button.connect("clicked", self.button_clicked, "1")
        hbox.pack_end(button, False, False, 10)
        hbox.show_all()

        vbox.pack_start(hbox, False, False, 0)

        #------------------------------------------------------
        first_opt_button = gtk.Button()
        opts = []
        if mode == "Clone":
            first_opt_button.set_label(_("Clone priority options"))
            opts = clone_priority_opts
        elif mode == "Restore":
            first_opt_button.set_label(_("Restore partitioning options"))
            opts = restore_partition_opts
        else:
            pass

        subVbox = gtk.VBox(False, 8)
        first_opt_button.connect("clicked", self.on_button_clicked, 0, subVbox)
        vbox.pack_start(first_opt_button, False, True, 0)

        #view = gtk.Viewport()
        #view.set_size_request(-1,300)
        #view.set_shadow_type(gtk.SHADOW_NONE)
        #view.add(subVbox)

        #scroll = gtk.ScrolledWindow()
        #scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        #scroll.set_shadow_type(gtk.SHADOW_IN)
        #scroll.add(view)

        vbox.pack_start(subVbox, False, False, 0)
        self.opt_stack.append(subVbox)

        if mode != "Onthefly":
            first_opt_button.show()

            first_radiobutton = None
            first = True
            index = 0
            for opt in opts:
                radioButton = gtk.RadioButton(first_radiobutton, opt[1])
                if opt[2] == True:
                    radioButton.set_active(True)

                if first:
                    first_radiobutton = radioButton
                    first = False

                #button.unset_flags(gtk.CAN_FOCUS)
                radioButton.connect("toggled", self.on_radiobutton_toggled, opts, index)
                radioButton.show()
                subVbox.pack_start(radioButton,True,False,2)

                index += 1

            separator = gtk.HSeparator()
            vbox.pack_start(separator, False, False, 0)

        #------------------------------------------------------
        mess_opt_button = gtk.Button()
        opts = []
        if mode == "Clone":
            mess_opt_button.set_label(_("Clone messive options"))
            opts = clone_mess_opts
        elif mode == "Restore":
            mess_opt_button.set_label(_("Restore advanced options"))
            opts = restore_adv_opts
        else:
            mess_opt_button.set_label(_("Onthefly advanced options 1"))
            opts = onthefly_adv_opts1

        subVbox = gtk.VBox(False, 8)
        mess_opt_button.connect("clicked", self.on_button_clicked, 1, subVbox)
        vbox.pack_start(mess_opt_button, False, True, 0)
        mess_opt_button.show()

        vbox.pack_start(subVbox, False, False, 0)
        self.opt_stack.append(subVbox)

        index = 0
        for opt in opts:
            button = gtk.CheckButton(opt[1])
            if opt[2] == True:
                button.set_active(True)

            button.connect("toggled",  self.on_radiobutton_toggled, opts, index)
            button.show()
            subVbox.pack_start(button,True,False,2)

        separator = gtk.HSeparator()
        vbox.pack_start(separator, False, False, 0)

        #------------------------------------------------------
        if mode != "Restore":
            adv_button = gtk.Button()
            opts = []
            if mode == "Clone":
                adv_button = gtk.Button(_("Clone compression options"))
                opts = clone_compression_opts
            else:
                adv_button = gtk.Button(_("Onthefly advanced options 2"))
                opts = onthefly_adv_opts2

            subVbox = gtk.VBox(False, 6)
            adv_button.connect("clicked", self.on_button_clicked, 2, subVbox)
            vbox.pack_start(adv_button, False, True, 0)
            adv_button.show()

            vbox.pack_start(subVbox, False, False, 0)
            self.opt_stack.append(subVbox)

            first_radiobutton = None
            first = True
            index = 0
            for opt in opts:
                radioButton = gtk.RadioButton(first_radiobutton, opt[1])
                if opt[2] == True:
                    radioButton.set_active(True)

                if first:
                    first_radiobutton = radioButton
                    first = False

                #button.unset_flags(gtk.CAN_FOCUS)
                radioButton.connect("toggled",  self.on_radiobutton_toggled, clone_compression_opts, index)
                radioButton.show()
                subVbox.pack_start(radioButton,True,False,2)

            separator = gtk.HSeparator()
            vbox.pack_start(separator, False, False, 0)
        #------------------------------------------------------
        vbox.show()
        return vbox

pass

#===============================================

class Stage5_Executing_Clonezilla_Core:
    def __init__(self, win, parent, cmd):
        self.win = win
        self.parent = parent
        self.cmd = cmd
        self.title = _("Processing")
        self.process_finished = False

        gobject.threads_init()
        self.worker = WorkerThread(self, self.cmd)

        self.view = gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        vbox = gtk.VBox(False, 5)
        vbox.show()
        self.view.add(vbox)

        #label =gtk.Label(text)
        #label.show()
        #vbox.pack_start(label, False, False, 10)

        hbox = gtk.HBox(False, 2)
        vbox.pack_end(hbox, False, False, 0)

        self.progressbar = gtk.ProgressBar()
        self.progressbar.set_fraction(0)
        vbox.pack_end(self.progressbar, False, True, 10)
        vbox.show_all()

        hbox = gtk.HBox(False, 2)
        vbox.pack_end(hbox, False, False, 0)

        self.msg_log = gtk.Label()
        self.msg_log.set_line_wrap(True)
        self.msg_log.show()
        hbox.pack_start(self.msg_log, False, False, 10)

        self.view.show_all()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def run_process(self):
        self.worker.start()

    def default_button_Setting(self, parent):
        parent.set_button(button="forward",mode="show")
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="backward",mode="show")
        parent.set_button(button="backward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")

    def go_next_page(self, parent):
        parent.set_button(button="forward",mode="inactive")
        parent.set_button(button="backward",mode="inactive")
        parent.set_button(button="cancel",mode="inactive")

        return Stage6_Finish_Executions_And_Show_Result(self.win)

pass

#===============================================

class Stage6_Finish_Executions_And_Show_Result:
    def __init__(self, win):
        self.win = win
        self.title = _("Job done!")
        self.process_finished = True

        self.view = gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        vbox = gtk.VBox(False, 5)
        hbox = gtk.HBox(False, 2)

        image = gtk.Image()
        image.set_from_pixbuf(gtk.gdk.pixbuf_new_from_file('pics/emblem-default.png'))
        image.show()
        hbox.pack_start(image, False, False, 10)

        hbox = gtk.HBox(False, 2)
        hbox.show_all()

        vbox.pack_start(hbox, False, False, 0)
        vbox.show_all()

        self.view.add(vbox)
        self.view.show_all()

    def get_widget(self):
        return self.view

    def get_title(self):
        return self.title

    def default_button_Setting(self,parent):
        parent.set_button(button="forward",mode="hide")
        parent.set_button(button="backward",mode="hide")
        parent.set_button(button="cancel",mode="active")
pass

#===============================================

class Select_Source_or_Target_Page:
    def __init__(self, win, stage, parent):
        self.win = win
        self.selection = ""
        self.stage = stage
        self.parent = parent
        self.tree = None

        self.device_selected = True

        self.view = gtk.Viewport()
        self.view.set_shadow_type(gtk.SHADOW_NONE)

        vbox = gtk.VBox(False, 5)
        vbox.show()
        self.view.add(vbox)

        image = gtk.Image()
        image.set_padding(0, 0)
        icon = gtk.STOCK_INFO
        image.set_from_stock(icon, gtk.ICON_SIZE_BUTTON)
        image.show()
        button = gtk.Button()
        button.add(image)
        button.set_relief(gtk.RELIEF_NONE)
        button.show()
        #button.connect("clicked", self.button_clicked, "1")

        self.hbox2 = gtk.HBox(False, 3)
        self.hbox2.pack_end(button, False, False, 0)
        vbox.pack_start(self.hbox2, False, False, 0)

        #===============================================
        self.fixed = gtk.Fixed()
        self.scroll = gtk.ScrolledWindow()
        self.scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        self.scroll.set_shadow_type(gtk.SHADOW_IN)
        self.scroll.show_all()

        self.fixed.put(self.scroll, 20, 0)
        self.fixed.show()

        vbox.pack_start(self.fixed, False, True, 10)

        #===============================================
        hbox = gtk.HBox(False, 2)

        btn=gtk.Button(stock=gtk.STOCK_REFRESH)
        btn.connect('clicked', self.get_select)
        btn.show()
        self.refresh_btn=btn
        hbox.pack_start(self.refresh_btn, False, False, 10)
        vbox.pack_start(hbox, False, False, 0)

        hbox = gtk.HBox(False, 2)
        self.msg_icon = gtk.Image()
        self.msg_icon.show()
        hbox.pack_start(self.msg_icon, False, False, 10)

        self.label = gtk.Label()
        self.label.show()
        hbox.pack_start(self.label, False, False, 10)

        vbox.pack_start(hbox, False, False, 0)
        vbox.show_all()

        self.view.show()

    def set_radiobutton(self, radiobutton1, radiobutton2):
        self.RadioButton1 = gtk.RadioButton(None , _("Device or partition."))
        self.RadioButton1.connect('toggled', self.get_select)
        self.hbox2.pack_start(self.RadioButton1, False, True, 5)
        self.RadioButton2 = gtk.RadioButton(self.RadioButton1 , _("Image file."))
        self.RadioButton2.connect('toggled', self.get_select)
        self.hbox2.pack_start(self.RadioButton2, False, True, 5)

        if radiobutton1:
            self.RadioButton1.show()
            self.RadioButton1.set_active(True)
            self.device_selected = True

            self.tree = self.device_partition_treeview()
            selection = self.tree.get_selection()
            selection.connect('changed', self.on_selection_changed)  
            self.scroll.add(self.tree)

        if radiobutton2:
            self.RadioButton2.show()

            if not radiobutton1:
                self.RadioButton2.set_active(True)
                self.device_selected = False

                self.tree = self.file_system_treeview()
                selection = self.tree.get_selection()
                selection.connect('changed', self.on_selection_changed) 
                self.scroll.add(self.tree)

    def get_radiobutton(self):
        # If Radio button 1 toggled, return True.
        return self.RadioButton1.get_active()

    def get_page(self):
        return self.view

    def get_select(self, widget):
        select_device = self.RadioButton1.get_active()
        self.label.set_label("")
        self.parent.set_button(button="forward",mode="inactive")

        if select_device:
            self.device_selected = True
            subtree = self.scroll.get_child()
            if subtree != None:
                self.scroll.remove(subtree)

            self.tree = self.device_partition_treeview()
            self.tree.expand_all()
            selection = self.tree.get_selection()
            selection.connect('changed', self.on_selection_changed)
            self.scroll.add(self.tree)
        else:
            self.device_selected = False
            subtree = self.scroll.get_child()
            if subtree != None:
                self.scroll.remove(subtree)

            self.tree = self.file_system_treeview()
            self.tree.expand_all()
            selection = self.tree.get_selection()
            selection.connect('changed', self.on_selection_changed) 
            self.scroll.add(self.tree)

    def on_selection_changed(self, selection): 
        (model, paths) = selection.get_selected()

        if paths:
            self.parent.set_button(button="forward",mode="active")
            self.parent.set_button(button="forward",mode="show")

            if self.device_selected:
                if len(model.get_path(paths)) == 1:
                    self.selection = "DEV::%s" % model.get_value(paths, 1)
                else:
                    self.selection = "PART::%s" % model.get_value(paths, 1)
                    if if_mounted(model.get_value(paths, 1)):
                        self.label.set_label("WARNING: This partition may be mounted!!")
                    else:
                        self.label.set_label(model.get_value(paths, 1))
            else:
                if len(model.get_path(paths)) > 2:
                    label = "%s%s" % ( _("Your selection is /home/partimag/"),\
                                       model.get_value(paths, 1) )
                else:
                    label = "%s%s" % ( _("Your selection is "),\
                                       model.get_value(paths, 1) )
                self.label.set_label(label)

    def get_selection(self):
        return self.selection

    def device_partition_treeview(self): 
        self.treestore = self.get_device_partition_list(All_Arguments['Source'])

        tree = gtk.TreeView(self.treestore)
        tree.connect('button_press_event', self.on_treeview_button_press_event)

        tree.set_size_request(400, 200)
        tree.set_headers_visible(False)
        tree.show()

        tvcolumn = gtk.TreeViewColumn('Column 0')
        tree.append_column(tvcolumn)

        cell = gtk.CellRendererPixbuf()         
        tvcolumn.pack_start(cell, False)

        cell2 = gtk.CellRendererText()
        tvcolumn.pack_start(cell2, True)

        cell3 = gtk.CellRendererText()
        tvcolumn.pack_start(cell3, True)

        cell4 = gtk.CellRendererText()
        tvcolumn.pack_start(cell4, True)

        tvcolumn.add_attribute(cell, 'pixbuf', 0)
        tvcolumn.add_attribute(cell2, 'text', 1)
        tvcolumn.add_attribute(cell3, 'text', 2)
        tvcolumn.add_attribute(cell4, 'text', 3)

        return tree

    def file_system_treeview(self):
        self.treestore = self.get_file_system_list()

        tree = gtk.TreeView(self.treestore)
        #tree.connect('button_press_event', self.on_treeview_button_press_event)
        tree.connect('row-activated', self.on_item_activated)

        tree.set_size_request(400, 200)
        tree.set_headers_visible(False)
        tree.show_all()

        tvcolumn = gtk.TreeViewColumn('Column 0')
        tree.append_column(tvcolumn)

        cell = gtk.CellRendererPixbuf()         
        tvcolumn.pack_start(cell, False)

        cell2 = gtk.CellRendererText()
        tvcolumn.pack_start(cell2, True)

        tvcolumn.add_attribute(cell, 'pixbuf', 0)
        tvcolumn.add_attribute(cell2, 'text', 1)
        return tree

    def get_device_partition_list(self,exception):
        excluded = ""
        if exception != "":
            excluded = exception.split("::")[1]

        treestore = gtk.TreeStore(gtk.gdk.Pixbuf, str,str,str)

        # To list the info. of all devices and partitions.
        sections = list_info_of_device_and_partition()

        for division in sections:            # [:-1] :: -1 for extra empty line.
            if division != "":
                div_lines = []
                if string.find(division, "BYT") == -1:
                    break
                div_lines = division.split(';')
                device = div_lines[1].split(':')

                if excluded != device[0].strip():
                    parent = treestore.append(None, [ gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), 
                                                      device[0].strip(), device[1].strip(), ""])

                    for partitions in div_lines[2:-1]:    # [2:-1] :: -1 for extra empty line.
                        items = partitions.split(':')

                        # partitionId may be like /dev/sda1, /dev/hda2, etc.
                        partitionId = "%s%s" % (device[0].strip(), items[0].strip())
                        if blkid_vaild(partitionId) and excluded != partitionId:
                            if string.find(items[4], "swap") == -1:
                                child = treestore.append(parent, [ gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), 
                                                                   partitionId, items[3].strip(), items[4].strip()])
        return treestore

    def get_file_system_list(self):
        treestore = gtk.TreeStore(gtk.gdk.Pixbuf, str)

        parent = treestore.append(None, [gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), _('Local Dev')])
        treestore.append(None, [gtk.gdk.pixbuf_new_from_file('pics/folder-remote.png'), _('SSH Server')])
        treestore.append(None, [gtk.gdk.pixbuf_new_from_file('pics/folder-remote.png'), _('FTP Server')])
        treestore.append(None, [gtk.gdk.pixbuf_new_from_file('pics/folder-remote.png'), _('NFS Server')])

        #  To list all the block devices except the SWAP partition. 
        all_partition = list_block_devices_except_SWAP()

        #  To list all the mounted file systems. 
        mounted_part = list_mounted_partition()

        #  To list all the unmounted partitions and /home/partimag if mounted. 
        unmounted = all_partition



        home_partimag_mount = None
        #for partition in all_partition:    # from blkid, list all block devices
        #    if_exist = False
        #    partitionId = partition[0]
        #    for mdpt in mounted_part:      # from mount, list all mounted partitions
        #        mdptId = mdpt[0].strip()
        #        mdptPath = mdpt[1].strip()
        #        if partitionId == mdptId:
        #            if mdptPath == "/home/partimag":
        #                home_partimag_mount = partition
        #            else:
        #                if_exist = True

        #    if if_exist == False:
        #        unmounted.append(partition)
        #==============================================================
        for umdpt in unmounted:
            if home_partimag_mount == None:
                treestore.append(parent, [ gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), 
                                                umdpt[0].strip()])
            else:
                if umdpt[0].strip() == home_partimag_mount[0].strip():
                    title = "%s %s" % (umdpt[0].strip(), " (/home/partimag)")
                    child = treestore.append(parent, [ gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), 
                                                            title])
                    for filename in os.listdir("/home/partimag"):
                        if os.path.isdir(os.path.join("/home/partimag", filename)):
                            treestore.append(child, [ gtk.gdk.pixbuf_new_from_file('pics/folder.png'), 
                                                           filename])
                else:
                    treestore.append(parent, [ gtk.gdk.pixbuf_new_from_file('pics/drive-harddisk.png'), 
                                                    umdpt[0].strip()])
        return treestore



    def on_item_activated(self, widget, row, col):
        #model = widget.get_model()
        if row[0] == 0:
            (model, paths) = widget.get_selection().get_selected()
            if len(model.get_path(paths)) > 1:
                dir_path = []

                # length == 2 means the level of partitions.
                if len(model.get_path(paths)) == 2:
                    if if_mounted(model.get_value(paths, 1)):
                        dir_path = list_child_dir(get_mounted_point(model.get_value(paths, 1)))
                    else:
                        tmp_dir = os.path.join("/tmp", get_random_string())
                        if not os.path.exists(tmp_dir):
                            os.mkdir(tmp_dir)
                        r = commands.getoutput("sudo mount -f %s %s" % (model.get_value(paths, 1),tmp_dir))
                        dir_path = list_child_dir(tmp_dir)
                else:
                    dir_path = []

                for child in dir_path:
                    # model here is a treestore object mapping to the treeview control.
                    model.append(paths ,[ gtk.gdk.pixbuf_new_from_file('pics/folder.png'), child])
                widget.expand_row(model.get_path(paths),False)


            #if len(model.get_path(paths)) < 3:
                #text = _("Are you sure to mount %s as /home/partimag to sore or read image files? ") % \
                #       model.get_value(paths, 1)

                #dialog = gtk.Dialog( "",
                #                     self.win,
                #                     gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,                       
                #                     (gtk.STOCK_OK, gtk.RESPONSE_YES, gtk.STOCK_CANCEL, gtk.RESPONSE_NO))
                #dialog.vbox.pack_start(gtk.Label(text),True,True,15)
                #dialog.show_all()
                #result = dialog.run()
                #if result == gtk.RESPONSE_YES:
                #    dialog.destroy()
                #    if not os.path.exists("/home/partimag"):
                #        try:
                #            os.makedirs("/home/partimag")
                #        except OSError:
                #            print str(OSError)
                #    elif os.path.ismount("/home/partimag"):
                #        try:
                        # """Subprocess may cause the process hang problem, so use commands instead currently. """
                        # subprocess.Popen(["sudo", "umount", "-f", "/home/partimag" ], stdout=subprocess.PIPE)
                #            r = commands.getoutput("sudo umount -f /home/partimag")
                #        except OSError:
                #            print str(OSError)

                    # """Subprocess may cause the process hang problem, so use commands instead currently. """
                    # p = subprocess.Popen(["mount", model.get_value(paths, 1), "/home/partimag" ], stdout=subprocess.PIPE)
                #    commands.getoutput("sudo mount %s /home/partimag" % model.get_value(paths, 1))
                #    if self.stage == "Select Source":
                #        self.refresh_btn.clicked()
                #    else:
                #        dialog2 = gtk.MessageDialog( self.win,  
                #                                     gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                #                                     gtk.MESSAGE_QUESTION,
                #                                     gtk.BUTTONS_OK, 
                #                                     None)
                #        dialog2.set_markup('Please enter the <b>saving image file name.</b>:')
                #        now = datetime.datetime.now()
                #        img_name = "%d-%d-%d-%d-img" % (now.year, now.month, now.day, now.hour)

                #        entry = gtk.Entry()
                #        entry.set_text(img_name)
                #        entry.connect("activate", self.responseToDialog, dialog2, gtk.RESPONSE_OK)

                #        hbox = gtk.HBox()
                #        hbox.pack_start(gtk.Label("Name:"), False, 5, 5)
                #        dialog2.format_secondary_markup("Input the name of the storing image file. Press \"Enter\" to accept the defualt name.")
                #        hbox.pack_end(entry)
                #        dialog2.vbox.pack_end(hbox, True, True, 0)
                #        dialog2.show_all()
                #        dialog2.run()
                #        text = entry.get_text()
                #        dialog2.destroy()             
                #        self.selection = text

                #        if not os.path.exists(os.path.join("/home/partimage",text)):
                #            os.makedirs(os.path.join("/home/partimag",text))
                        
                #        self.refresh_btn.clicked()
                #else:
                #    dialog.destroy()
            return 0
        elif row[0] == 1:
            text = _("SSH server connectivity is not supported yet, we will work harder to make it coming up soon!")
            md = gtk.MessageDialog(self.win, 
                                   gtk.DIALOG_DESTROY_WITH_PARENT, 
                                   gtk.MESSAGE_WARNING, 
                                   gtk.BUTTONS_CLOSE, 
                                   text )
            md.run()
            md.destroy()
        elif row[0] == 2:
            text = _("FTP server connectivity is not supported yet, we will work harder to make it coming up soon!")
            md = gtk.MessageDialog(self.win, 
                                   gtk.DIALOG_DESTROY_WITH_PARENT, 
                                   gtk.MESSAGE_WARNING, 
                                   gtk.BUTTONS_CLOSE, 
                                   text )
            md.run()
            md.destroy()
        else:
            text = _("NFS server connectivity is not supported yet, we will work harder to make it coming up soon!")
            md = gtk.MessageDialog(self.win, 
                                   gtk.DIALOG_DESTROY_WITH_PARENT, 
                                   gtk.MESSAGE_WARNING, 
                                   gtk.BUTTONS_CLOSE, 
                                   text )
            md.run()
            md.destroy()


    def responseToDialog(entry, dialog, response):
        dialog.response(response)


    #http://faq.pygtk.org/index.py?query=clicks&querytype=simple&req=search
    #http://faq.pygtk.org/index.py?req=show&file=faq05.004.htp
    def on_treeview_button_press_event(self, treeview, event):
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            #time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                self.contextMenu(treeview, event)
                #self.popup.popup( None, None, None, event.button, time)
            return True


    def contextMenu(self, widget, event, data=None):
        menu = gtk.Menu()
        one = gtk.MenuItem("Mount as /home/partimag")
        #one.connect("activate", self.hello)
        menu.append(one)

        two = gtk.MenuItem("Umount")
        #two.connect("activate", self.hello)
        menu.append(two)

        menu.show_all()
        menu.popup(None, None, None, event.button, event.get_time())
        return True

class WorkerThread(threading.Thread):
    def __init__(self, parent, cmd):
        threading.Thread.__init__(self)
        self.parent = parent
        self.cmd = cmd
        self.mode = All_Arguments['Operation']
        #self.total = 0

    def run(self):
        self.work_complete = False
        self.amount_completed = 0
        if self.mode == "onthefly": 
            gobject.timeout_add(100, self._update_bar_onthefly)
        elif self.mode == "savedisk" or self.mode == "saveparts":
            gobject.timeout_add(100, self._update_bar_clone)
        else:
            gobject.timeout_add(100, self._update_bar_restore)

        count = 0
        run_cmd = "yes | LANG=en.US "
        run_cmd += self.cmd
        run_cmd += " 2> /tmp/partclone_processbar.log&"

        print run_cmd
        p = subprocess.Popen(run_cmd, shell=True, stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)

        text = p.stdout.readline()
        while text:
            if string.find(text, "***") == -1:
                tmp_list = text.split(";39m")
                if len(tmp_list) > 1:
                    text = tmp_list[-1]
                else:
                    tmp_list2 = text.split(";33m")
                    if len(tmp_list2) > 1:
                        text = tmp_list2[-1]
                    else:
                        text = tmp_list2[0]

                if text.find("Is the disk full") != -1:  # for savedisk/parts case
                    self.parent.msg_log.set_text("Start cloning...")
                else:
                    self.parent.msg_log.set_text(text)

            text = p.stdout.readline()

        p.wait()
        self.work_complete = True
        self.parent.msg_log.set_text("Job done......")
        self.parent.parent.go_forward_btn.clicked()

    def _update_bar_onthefly(self):
        if os.path.exists("/tmp/partclone_processbar.log"):
            p = subprocess.Popen("tail -n 1 /tmp/partclone_processbar.log&", shell=True, stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            text = p.communicate()[0]
            textPattern = re.compile("Completed:\s+(?P<percent>[0-9]+\.[0-9]+\%)")
            #Elapsed: 00:01:31, Remaining: 00:00:14, Completed: 86.29%, Rate: 859.06MB/min,
            for t in re.findall(textPattern, text):
                s = t.split('%')[0]
                i = float(s)/100
                self.parent.progressbar.set_fraction(i)
                self.parent.progressbar.set_text("%s%%" %s)

            if text.find('successful') != -1:
                self.parent.progressbar.set_fraction(1.0)
                self.parent.progressbar.set_text("100%")

        return not self.work_complete

    def _update_bar_clone(self):
        if os.path.exists("/tmp/partclone_processbar.log"):
            p = subprocess.Popen("tail -n 1 /tmp/partclone_processbar.log&", shell=True, stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            text = p.communicate()[0]
            textPattern = re.compile("Completed:\s+(?P<percent>[0-9]+\.[0-9]+\%)")
            #Elapsed: 00:01:31, Remaining: 00:00:14, Completed: 86.29%, Rate: 859.06MB/min,
            for t in re.findall(textPattern, text):
                s = t.split('%')[0]
                i = float(s)/100
                self.parent.progressbar.set_fraction(i)
                self.parent.progressbar.set_text("%s%%" %s)

            if text.find('Finished') != -1:
                self.parent.progressbar.set_fraction(1.0)
                self.parent.progressbar.set_text("100%")

        return not self.work_complete

    def _update_bar_restore(self):
        if os.path.exists("/tmp/partclone_processbar.log"):
            p = subprocess.Popen("tail -n 1 /tmp/partclone_processbar.log&", shell=True, stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            text = p.communicate()[0]
            textPattern = re.compile("Completed:\s+(?P<percent>[0-9]+\.[0-9]+\%)")
            #Elapsed: 00:01:31, Remaining: 00:00:14, Completed: 86.29%, Rate: 859.06MB/min,
            for t in re.findall(textPattern, text):
                s = t.split('%')[0]
                i = float(s)/100
                self.parent.progressbar.set_fraction(i)
                self.parent.progressbar.set_text("%s%%" %s)

            if text.find('successful') != -1:
                self.parent.progressbar.set_fraction(1.0)
                self.parent.progressbar.set_text("100%")

        return not self.work_complete

