#!/usr/bin/env python
# -*- encoding : ascii -*-
#
# The main program of PyClonezilla
#
# Copyright (C) 2010~ Zong-shuo Jheng <c00aso00 at nchc org tw>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pygtk
pygtk.require('2.0')
import os, sys
import gtk, gobject
import shlex, subprocess
import locale, gettext
import re
import string
import random

""" How to re-generate a modified .mo file:
1. create Portable Object Template (POT) file:
   'xgettext --language=Python --keyword=_ --output=XXXX.pot  AAA.py BBB.py ....'

2. Create a PO file. The locale depended on which locale you want to generate.
   'msginit --input=XXXX.pot --locale=zh-TW'
   Edit the XXXX.pot file.

3. Create .mo file which is used for translation by gettext module.
   'mkdir -p locale/zh_TW/LC_MESSAGES'  if the locale dir does not exist.
   'msgfmt --output-file=locale/zh_TW/LC_MESSAGES/YOUR_APP_NAME.mo XXXX.po' """


APP_NAME="PyClonezilla"

## i18n
## Copy from "blog learning python" http://www.learningpython.com/2006/12/03/translating-your-pythonpygtk-application/
#Translation stuff

#Get the local directory since we are not installing anything
local_path = os.path.realpath(os.path.dirname(sys.argv[0]))
local_path = os.path.join (local_path, 'locale')

# Init the list of languages to support
langs = []

#Check the default locale
lc, encoding = locale.getdefaultlocale()
if (lc):
    #If we have a default, it's the first in the list
    langs = [lc]

# Now lets get all of the supported languages on the system
language = os.environ.get('LANGUAGE', None)
if (language):
    """langage comes back something like en_CA:en_US:en_GB:en
    on linuxy systems, on Win32 it's nothing, so we need to
    split it up into a list"""

    langs += language.split(":")

"""Now add on to the back of the list the translations that we
know that we have, our defaults"""
langs += ["en_CA", "en_US"]

"""Now langs is a list of all of the languages that we are going
to try to use.  First we check the default, then what the system
told us, and finally the 'known' list"""

gettext.bindtextdomain(APP_NAME, local_path)
gettext.textdomain(APP_NAME)

# Get the language to use
lang = gettext.translation(APP_NAME, local_path
                           , languages=langs, fallback = True)

"""Install the language, map _() (which we marked our
strings to translate with) to lang.gettext() which will
translate them."""
_ = lang.gettext

## i18n done

def parsing_df_result():
    p = subprocess.Popen(["df", "-h"], stdout=subprocess.PIPE)
    text = p.communicate()[0]

    textPattern = re.compile(r"(?P<device>/[a-z]+/[a-z]+[0-9]?)\s+"
                             r"(?P<amount>[0-9]+\.?[0-9]?[T|G|M|K]+)\s+"
                             r"(?P<available>[0-9]+\.?[0-9]?[T|G|M|K]+)\s+"
                             r"(?P<used>[0-9]+\.?[0-9]?[T|G|M|K]+)\s+"
                             r"(?P<usage>[0-9]+\.?[0-9]?%)\s+"
                             r"(?P<mountpoint>[^\s]+)")

    return re.findall(textPattern, text)     
    # Return tuples ::  tt[0]:<device> tt[1]:<amount>  tt[2]:<available>
    #                   tt[3]:<used>   tt[4]:<usage>   tt[5]:<mountpoint>
    # for tt in re.findall(textPattern, text):
    #        print tt[0]  tt[1]  tt[2]  tt[3]  tt[4]  tt[5]


        #for tt in text.splitlines(): #re.split(r'[^\n]a+', text):
        #    result = textPattern.search(tt)
        #    if result == None:
        #        pass
        #    else:
        #        print result.groups()

        

        #if result == None:
        #    print "Not found"
        #else:
            #print result.group('device')
            #print result.group('amount')
            #print result.group('available')
            #print result.group('used')
            #print result.group('usage')
            #print result.group('mountpoint')
            #print result.groups(0)

pass

def parsing_fdisk_result():
    p = subprocess.Popen(["fdisk", "-l"], stdout=subprocess.PIPE)
    text = p.communicate()[0]

    textPattern = re.compile(r"(?P<device>/[a-z]+/[a-z]+[0-9]?)\s+\*?\s+"
                             r"(?P<start>[0-9]+)\s+"
                             r"(?P<end>[0-9]+)\s+"
                             r"(?P<blocks>[0-9]+\+?)\s+"
                             r"(?P<id>[0-9a-z]+)\s+"
                             r"(?P<system>[^\n]+)")

    return re.findall(textPattern, text)     

pass

def parsing_parted_result():
    p = subprocess.Popen(["parted", "-lm"], stdout=subprocess.PIPE)
    text = p.communicate()[0]

    textPattern = re.compile(r"(?P<device>/[a-z]+/[a-z]+[0-9]?)\s+\*?\s+"
                             r"(?P<start>[0-9]+)\s+"
                             r"(?P<end>[0-9]+)\s+"
                             r"(?P<blocks>[0-9]+\+?)\s+"
                             r"(?P<id>[0-9a-z]+)\s+"
                             r"(?P<system>[^\n]+)")

    return re.findall(textPattern, text)

pass


# To ensure selected partitions to be in 'blkid' result.
def blkid_vaild(partitionId):
    p = subprocess.Popen(["blkid"], stdout=subprocess.PIPE)
    text = p.communicate()[0]
    string = []
    string = text.split('\n')

    for readline in string:
        part = readline.split(':')
        if partitionId == part[0]:
            return True
    return False
pass

# To check if the selected device or partition mounted or not.
def if_mounted(partitionId):
    p = subprocess.Popen(["mount"], stdout=subprocess.PIPE)
    text = p.communicate()[0]
    string = []
    string = text.split('\n')

    for readline in string:
        part = readline.split(' ')
        if partitionId == part[0]:
            return True
    return False
pass

# To get the mounted point from partition id.
def get_mounted_point(partitionId):
    for part in list_mounted_partition():
        mdptId = part[0].strip()
        mdptPath = part[1].strip()
        if mdptId == partitionId:
            return mdptPath
    return ""
pass


# To list all the block devices except the SWAP partition.
def list_block_devices_except_SWAP():
    p = subprocess.Popen(["blkid"], stdout=subprocess.PIPE)
    text = p.communicate()[0]

    temp = []
    temp = text.split('\n')

    all_partition = []
    for part in temp:                             # [:-1] :: -1 for extra empty line.
        if part != "":
            if string.find(part, "TYPE=\"swap\"") == -1:   # To filter the SWAP partition out.
                all_partition.append(part.split(':'))

    return all_partition
pass


# To list all the mounted file systems. 
def list_mounted_partition():
    p = subprocess.Popen(["mount"], stdout=subprocess.PIPE)
    text = p.communicate()[0]
    textPattern = re.compile(r"(?P<device>/[a-z]+/[a-z]+[0-9]?)\s+on\s+"
                             r"(?P<mount_point>[^\s]+)\s+type\s+"
                             r"(?P<type>[^\s]+)\s+")
    temp = []
    temp = re.findall(textPattern, text)

    mounted_part = []
    for mp in temp:
        mounted_part.append(mp)

    return mounted_part
pass


# To list the info. of all devices and partitions.
def list_info_of_device_and_partition():
    p = subprocess.Popen("yes | parted -lm", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    text = p.communicate()[0]
    #==============================================================
    #  To get the info. of all devices and partitions.
    #  Returned example:
    #      BYT;
    #      /dev/sda:60.0GB:scsi:512:512:msdos:ATA ST96812AS;
    #      1:1049kB:57.5GB:57.5GB:ext4::;
    #      2:57.5GB:60.0GB:2496MB:::;
    #      5:57.5GB:60.0GB:2496MB:linux-swap(v1)::; 
    #==============================================================
    sections = []
    sections = text.split('\n\n')

    return sections
pass


# To list all the child directories.
def list_child_dir(path):
    if not os.path.exists(path):
        return null
    else: 
        dir_list = []
        for filename in os.listdir(path):
            if os.path.isdir(os.path.join(path, filename)):
                dir_list.append(filename)
        return dir_list
pass


# To list the info. of the selected device or partition.
def list_info_of_selection(selected):
    return ""
pass



# Generate random string.
def get_random_string():
    alphabet = 'abcdefghijklmnopqrstuvwxyz'
    min = 5
    max = 15
    total = 20
    string = ''
    for count in xrange(1,total):
        for x in random.sample(alphabet,random.randint(min,max)):
            string+=x
    return string


class CellRendererProgressBar(gtk.GenericCellRenderer):
    __gproperties__ = { "progressbar": (gobject.TYPE_OBJECT, 
                                        "Progress Bar", 
                                        "Customized progress bar for CellRenderer", 
                                        gobject.PARAM_READWRITE), }

    def __init__(self):
        self.__gobject_init__()
        self.progressbar = gtk.ProgressBar()

        return self.progressbar

    #def do_set_text(self, property, value):
    #    self.progressbar.set_text(value)

    #def do_set_property(self, pspec, value):
    #    setattr(self, pspec.name, value)

    #def do_get_property(self, pspec):
    #    return getattr(self, pspec.name)

    #def on_render(self): #, window, widget, background_area, cell_area, expose_area, flags):
    #    progressbar = gtk.ProgressBar()
    #    self.custom.draw(window, widget, cell_area.x, cell_area.y)

    #def on_get_size(self, widget, cell_area=None):
    #    return (0, 0, self.custom.get_width(), self.custom.get_height())

gobject.type_register(CellRendererProgressBar)

pass


